@@ -30,6 +30,10 @@ typedef struct nearest_module
     FILE * file;
     /* Is this file a gz file? */
     int is_gz : 1;
+    /* Print blah messages? */
+    int verbose : 1;
+    /* Actually found something? */
+    int found : 1;
     /* The term to search for. */
     const char * search_term;
     /* The length of the search term. */
@@ -61,28 +65,52 @@ static void fail (int test, const char * message, ...)
     return;
 }
 
-/* Copy-paste from
-   "http://www.lemoda.net/c/levenshtein/index.html". */
-
 static int distance (const char * word1,
                      int len1,
                      const char * word2,
-                     int len2)
+                     int len2,
+                     int max)
 {
     int matrix[len1 + 1][len2 + 1];
     int i;
+    int j;
+
     for (i = 0; i <= len1; i++) {
         matrix[i][0] = i;
     }
-    for (i = 0; i <= len2; i++) {
-        matrix[0][i] = i;
+    for (j = 0; j <= len2; j++) {
+        matrix[0][j] = j;
+    }
+    /* If the maximum is small enough to make any difference, fill in
+       every cell of the matrix with a "too big" number. */
+    if (max < len1 || max < len2) {
+        for (i = 1; i <= len1; i++) {
+            for (j = 1; j <= len2; j++) {
+                matrix[i][j] = max + 1;
+            }
+        }
     }
     for (i = 1; i <= len1; i++) {
-        int j;
         char c1;
+        /* The bottom of the column. */
+        int min_j;
+        /* The top of the column. */
+        int max_j;
+        /* The smallest value of the matrix in column i. */
+        int col_min;
 
+        min_j = 1;
+        if (i > max) {
+            min_j = i - max;
+        }
+        max_j = len2;
+        if (len2 > max + i) {
+            max_j = max + i;
+        }
+        //        printf ("%d - %d\n", min_j, max_j);
         c1 = word1[i-1];
-        for (j = 1; j <= len2; j++) {
+        col_min = INT_MAX;
+        for (j = min_j; j <= max_j; j++) {
             char c2;
 
             c2 = word2[j-1];
@@ -107,6 +135,21 @@ static int distance (const char * word1,
                 }
                 matrix[i][j] = minimum;
             }
+            if (matrix[i][j] < col_min) {
+                col_min = matrix[i][j];
+            }
+        }
+        if (col_min > max) {
+            /* All the elements of this column are greater than the
+               maximum. */
+            /*
+            printf ("All too big at column %d.\n", i);
+            for (j = min_j; j <= max_j; j++) {
+                printf ("%d:%d ", j, matrix[i][j]);
+            }
+            printf ("\n");
+            */
+            return max + 1;
         }
     }
     return matrix[len1][len2];
@@ -128,8 +171,14 @@ static void nearest_compare_line (nearest_module_t * nearest)
         ;
     b[l] = '\0';
 
-    d = distance (b, l, nearest->search_term, nearest->search_len);
+    d = distance (b, l, nearest->search_term, nearest->search_len,
+                  nearest->distance);
     if (d < nearest->distance) {
+        if (nearest->verbose) {
+            printf ("%s (%d) is nearer than %d.\n", nearest->buf, d,
+                    nearest->distance);
+        }
+        nearest->found = 1;
         nearest->distance = d;
         strncpy (nearest->nearest, nearest->buf, l);
         nearest->nearest[l] = '\0';
@@ -205,8 +254,13 @@ nearest_get_line (nearest_module_t * nearest)
 
 static void search_packages (nearest_module_t * nearest)
 {
+    static int max_sane_distance = 10;
     nearest_open_file (nearest);
-    nearest->distance = INT_MAX;
+    /* Don't use INT_MAX here or get overflow. */
+    nearest->distance = nearest->search_len + 1;
+    if (nearest->search_len > max_sane_distance) {
+        nearest->distance = max_sane_distance;
+    }
     while (nearest_get_line (nearest)) {
         nearest_compare_line (nearest);
     }
@@ -240,19 +294,29 @@ cpan_nearest_search (char * file_name, char * search_term)
     nearest_set_search_file (& nearest, file_name);
 
     search_packages (& nearest);
-    return strdup (nearest.nearest);
+    if (nearest.found) {
+        return strdup (nearest.nearest);
+    }
+    else {
+        return 0;
+    }
 }
 
 #ifdef TEST
 
 static void print_result (nearest_module_t * nearest)
 {
-    if (nearest->distance == 0) {
-        printf ("Found '%s'\n", nearest->nearest);
+    if (nearest->found) {
+        if (nearest->distance == 0) {
+            printf ("Found '%s'\n", nearest->nearest);
+        }
+        else {
+            printf ("Closest to '%s' is '%s'.\n", nearest->search_term,
+                    nearest->nearest);
+        }
     }
     else {
-        printf ("Closest to '%s' is '%s'.\n", nearest->search_term,
-                nearest->nearest);
+        printf ("Nothing similar was found.\n");
     }
     return;
 }
@@ -261,13 +325,18 @@ int main (int argc, char ** argv)
 {
     nearest_module_t nearest = {0};
     char * st;
-    
+    if (strcmp (argv[1], "-v") == 0) {
+        nearest.verbose = 1;
+        st = argv[2];
+    }
+    else {
     if (argc < 2) {
         st = "Lingua::Stop::Weirds";
     }
     else {
         st = argv[1];
     }
+    }
     nearest_set_search_term (& nearest, st);
     nearest_set_search_file (& nearest, "02packages.details.txt");
 
